"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const PREFIX = 'connection:';
exports.base64 = (str) => new Buffer(str, 'ascii').toString('base64');
exports.unbase64 = (b64) => new Buffer(b64, 'base64').toString('ascii');
/**
 * Rederives the offset from the cursor string
 */
function cursorToOffset(cursor) {
    return parseInt(exports.unbase64(cursor).substring(PREFIX.length), 10);
}
exports.cursorToOffset = cursorToOffset;
/**
 * Given an optional cursor and a default offset, returns the offset to use;
 * if the cursor contains a valid offset, that will be used, otherwise it will
 * be the default.
 */
function getOffsetWithDefault(cursor, defaultOffset) {
    if (cursor === undefined) {
        return defaultOffset;
    }
    const offset = cursorToOffset(cursor);
    return isNaN(offset) ? defaultOffset : offset;
}
exports.getOffsetWithDefault = getOffsetWithDefault;
/**
 * Creates the cursor string from an offset.
 */
function offsetToCursor(offset) {
    return exports.base64(PREFIX + offset);
}
exports.offsetToCursor = offsetToCursor;
/**
 * Accepts a mongoose query and connection arguments, and returns a connection
 * object for use in GraphQL. It uses array offsets as pagination, so pagination
 * will work only if the data set is static.
 */
function paginate(rootValue, args, context) {
    return __awaiter(this, void 0, void 0, function* () {
        const { mapper, repository, queryBuilder = repository.createQueryBuilder() } = context;
        const { after, before, first, last, sort } = args;
        const totalCount = yield repository.count();
        const beforeOffset = getOffsetWithDefault(before, totalCount);
        const afterOffset = getOffsetWithDefault(after, -1);
        let startOffset = Math.max(-1, afterOffset) + 1;
        let endOffset = Math.min(totalCount, beforeOffset);
        if (first) {
            endOffset = Math.min(endOffset, startOffset + first);
        }
        if (last) {
            startOffset = Math.max(startOffset, endOffset - last);
        }
        const skip = Math.max(startOffset, 0);
        const limit = endOffset - startOffset;
        if (Array.isArray(sort)) {
            const typeOrmSort = sort.reduce((sortObject, s) => (Object.assign({}, sortObject, { [s.field]: s.order.toLowerCase() })), {});
            queryBuilder.orderBy(typeOrmSort);
        }
        // If supplied slice is too large, trim it down before mapping over it.
        queryBuilder.skip(skip).take(limit);
        // Short circuit if limit is 0; in that case, mongodb doesn't limit at all
        let slice = [];
        let count = 0;
        if (limit > 0) {
            [slice, count] = yield queryBuilder.getManyAndCount();
        }
        // If we have a mapper function, map it!
        if (typeof mapper === 'function') {
            slice = slice.map(mapper);
        }
        const edges = slice.map((value, index) => ({
            cursor: offsetToCursor(startOffset + index),
            node: value,
        }));
        const firstEdge = edges[0];
        const lastEdge = edges[edges.length - 1];
        const lowerBound = after ? afterOffset + 1 : 0;
        const upperBound = before ? Math.min(beforeOffset, totalCount) : totalCount;
        return {
            edges,
            pageInfo: {
                startCursor: firstEdge ? firstEdge.cursor : null,
                endCursor: lastEdge ? lastEdge.cursor : null,
                hasPreviousPage: last !== null ? startOffset > lowerBound : false,
                hasNextPage: first !== null ? endOffset < upperBound : false,
                count,
            },
        };
    });
}
exports.paginate = paginate;
//# sourceMappingURL=paginate.js.map